import numpy as np
import pandas as pd


def calculate_coverage(depth_file, threshold, edge=75, is_combined=False,
                       contig_threshold=1000, sep=None, binned_thre_dict=None):
    """
    Input is position depth file generated from mosdepth or bedtools genomecov
    """
    data_contig_list = []
    coverage = []
    var = []

    split_contig_list = []
    split_coverage = []
    f = open(depth_file)

    def process(depth_value, contig_name):
        depth_value_ = depth_value[edge:len(depth_value) - edge]
        mean = np.mean(depth_value_)
        coverage.append(mean)
        var.append(np.var(depth_value_))
        data_contig_list.append(contig_name)
        if is_combined:
            if len(depth_value) >= threshold:
                depth_value1 = depth_value[0:len(depth_value) // 2]
                depth_value2 = depth_value[len(
                    depth_value) // 2: len(depth_value)]
                split_contig_list.append(contig_name + '_1')
                split_contig_list.append(contig_name + '_2')
                depth_value1 = depth_value1[edge:len(depth_value1)]
                mean = np.mean(depth_value1)
                split_coverage.append(mean)
                depth_value2 = depth_value2[0:len(depth_value2) - edge]
                mean = np.mean(depth_value2)
                split_coverage.append(mean)

    depth_value = []
    depth_contig = None

    for line in f:
        line_split = line.strip().split('\t')
        contig_name = line_split[0]
        length = int(line_split[2]) - int(line_split[1])
        value = int(line_split[3])

        if depth_contig is None:
            depth_contig = contig_name
        if contig_name == depth_contig:
            depth_value.extend([value] * length)
        if contig_name != depth_contig:
            if sep is None:
                cov_threshold = contig_threshold
            else:
                sample_name = depth_contig.split(sep)[0]
                cov_threshold = 1000 if sample_name in binned_thre_dict[1000] else 2500

            if len(depth_value) <= cov_threshold:
                depth_contig = contig_name
                depth_value = []
                depth_value.extend([value] * length)
                continue

            process(depth_value, depth_contig)
            depth_contig = contig_name
            depth_value = []
            depth_value.extend([value] * length)

    if depth_value != []:
        if sep is None:
            cov_threshold = contig_threshold
        else:
            sample_name = depth_contig.split(sep)[0]
            cov_threshold = 1000 if sample_name in binned_thre_dict[1000] else 2500
        if len(depth_value) > cov_threshold:
            process(depth_value, depth_contig)

    if is_combined:
        data_contig_list = np.expand_dims(np.array(data_contig_list), axis=1)
        coverage = np.expand_dims(np.array(coverage), axis=1)
        contig_cov = np.concatenate((data_contig_list, coverage), axis=1)
        contig_cov = pd.DataFrame(
            data=contig_cov[:, 1:], index=contig_cov[:, 0], columns=['cov'])
        contig_cov['cov'] = contig_cov['cov'].astype('float')
        split_contig_list = np.expand_dims(np.array(split_contig_list), axis=1)
        split_coverage = np.expand_dims(np.array(split_coverage), axis=1)
        split_contig_cov = np.concatenate(
            (split_contig_list, split_coverage), axis=1)
        split_contig_cov = pd.DataFrame(
            data=split_contig_cov[:, 1:], index=split_contig_cov[:, 0], columns=['cov'])
        split_contig_cov['cov'] = split_contig_cov['cov'].astype('float')
        return contig_cov, split_contig_cov
    else:
        data_contig_list = np.expand_dims(np.array(data_contig_list), axis=1)
        coverage = np.expand_dims(np.array(coverage), axis=1)
        var = np.expand_dims(np.array(var), axis=1)
        contig_cov = np.concatenate((data_contig_list, coverage), axis=1)
        contig_cov = np.concatenate((contig_cov, var), axis=1)
        contig_cov = pd.DataFrame(
            data=contig_cov[:, 1:], index=contig_cov[:, 0], columns=['mean', 'var'])
        contig_cov['mean'] = contig_cov['mean'].astype('float')
        contig_cov['var'] = contig_cov['var'].astype('float')
        return contig_cov



import argparse
import sys
import logging
import os
from .utils import validate_args, get_threshold, generate_cannot_link
import subprocess
import gzip
import bz2
from Bio import SeqIO
import multiprocessing
from .generate_coverage import calculate_coverage
from atomicwrites import atomic_write
from .generate_kmer import generate_kmer_features_from_fasta
import pandas as pd
from Bio.SeqRecord import SeqRecord
from .semi_supervised_model import train
import torch
from .cluster import cluster
import shutil

def parse_args(args):
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description='Semi-supervised siamese neural network for metagenomic binning')

    subparsers = parser.add_subparsers(title='S3N2Bin subcommands',
                                       dest='cmd',
                                       metavar='')

    single_easy_bin = subparsers.add_parser('single_easy_bin',
                                            help='Get the bins with single or co-assembly binning using one line command.')

    multi_easy_bin = subparsers.add_parser('multi_easy_bin',
                                            help='Get the bins with multi-samples binning using one line command.')

    predict_taxonomy = subparsers.add_parser('predict_taxonomy',
                                             help='Run the contig annotation using mmseqs '
                                                  'with GTDB reference genome and generate'
                                                  'cannot-link file used in the semi-supervsied deep learning model training.'
                                                  '(Will download the GTDB database if not input path of GTDB )')

    generate_data_single = subparsers.add_parser('generate_data_single',
                                                 help='Generate training data(data.csv,data_split.csv) '
                                                      'for single and co-assembly binning '
                                                      'for the semi-supervised deep learning model training.')

    generate_data_multi = subparsers.add_parser('generate_data_multi', help='Generate training data(data.csv,data_split.csv) '
                                          'for multi-samples binning '
                                          'for the semi-supervised deep learning model training.')

    binning = subparsers.add_parser('bin',
                                    help='Training the model and clustering contigs to bins.')

    binning.add_argument('--data',
                         required=True,
                         help='Path to the input data.csv file.',
                         dest='data',
                         default=None,
                         )
    binning.add_argument('--data-split',
                         required=True,
                         help='Path to the input data_split.csv file.',
                         dest='data_split',
                         default=None,
                         )
    binning.add_argument('-c', '--cannot-link',
                         required=True,
                         nargs='*',
                         help='Path to the input cannot link file generated from other additional biological information,'
                         'one row for each cannot link constraint.'
                         'The file format: contig_1,contig_2.',
                         dest='cannot_link',
                         default=None,
                         metavar='')

    for p in [single_easy_bin, multi_easy_bin, predict_taxonomy, generate_data_single, generate_data_multi,binning]:
        p.add_argument('-i', '--input-fasta',
                                required=True,
                                help='Path to the input fasta file.',
                                dest='contig_fasta',
                                default=None,)

    for p in [single_easy_bin, multi_easy_bin, predict_taxonomy, generate_data_single, generate_data_multi, binning]:
        p.add_argument('-o', '--output',
                            required=True,
                            help='Output directory (will be created if non-existent)',
                            dest='output',
                            default=None,
                            )

    for p in [single_easy_bin, multi_easy_bin, generate_data_single, generate_data_multi, binning]:
        p.add_argument('-b', '--input-bam',
                            required=True,
                            nargs='*',
                            help='Path to the input BAM file. '
                                 'If using multiple sample binning, you can input multiple files.',
                            dest='bams',
                            default=None,
                            )

        p.add_argument('-p', '--processes', '-t', '--threads',
                                     required=False,
                                     type=int,
                                     help='Number of CPUs used (pass the value 0 to use all CPUs)',
                                     dest='num_process',
                                     default=0,
                                     metavar=''
                                     )

    for p in [single_easy_bin, multi_easy_bin, predict_taxonomy]:
        p.add_argument('-r', '--reference-db',
                            required=False,
                            help='GTDB reference file. (Default: $HOME/.cache/S3N2Bin/mmseqs2-GTDB/GTDB).'
                            'If not set --reference-db and can not find GTDB in $HOME/.cache/S3N2Bin/mmseqs2-GTDB/GTDB, '
                            'we will download GTDB to the default path.',
                            dest='GTDB_reference',
                            metavar='',
                            default=None)
    for p in [single_easy_bin, predict_taxonomy]:
        p.add_argument('--cannot-name',
                            required=False,
                            help='Name for the cannot-link file.',
                            dest='cannot_name',
                            default='cannot',
                            metavar=''
                            )

    for p in [single_easy_bin, multi_easy_bin, binning]:
        p.add_argument('--minfasta-kbs',
                            required=False,
                            type=int,
                            help='minimum bin size in Kbps (Default: 200).',
                            dest='minfasta_kb',
                            default=200,
                            metavar='')

        p.add_argument('--epoches',
                          required=False,
                          type=int,
                          help='Number of epoches used in the training process (Default: 20).',
                          dest='epoches',
                          default=20)

        p.add_argument('--batch-size',
                          required=False,
                          type=int,
                          help='Batch size used in the training process (Default: 2048).',
                          dest='batchsize',
                          default=2048,)

        p.add_argument('--max-edges',
                          required=False,
                          type=int,
                          help='The maximum number of edges that can be connected to one contig (Default: 200).',
                          dest='max_edges',
                          default=200)

        p.add_argument('--max-node',
                          required=False,
                          type=float,
                          dest='max_node',
                          default=1,
                          help='Fraction of contigs that considered to be binned (should be between 0 and 1; default: 1).')

    for p in [multi_easy_bin, generate_data_multi]:
        p.add_argument('-s', '--separator',
                           required=False,
                           type=str,
                           help='Used when multiple samples binning to separate sample name and contig name (Default is :).',
                           dest='separator',
                           default=':',
                           metavar='')


    if not args:
        parser.print_help(sys.stderr)
        sys.exit()
    return parser.parse_args(args)


def generate_cov(bam_file, bam_index, out, threshold,
                 is_combined, contig_threshold, logger):
    logger.info('Processing {}'.format(bam_file))
    bam_name = os.path.split(bam_file)[-1] + '_{}'.format(bam_index)
    bam_depth = os.path.join(out, '{}_depth.txt'.format(bam_name))
    os.system(
        'bedtools genomecov -bga -ibam {0} > {1}'.format(bam_file, bam_depth))

    if is_combined:
        contig_cov, must_link_contig_cov = calculate_coverage(bam_depth, threshold, is_combined=is_combined,
                                                              contig_threshold=contig_threshold)
        contig_cov = contig_cov.apply(lambda x: x + 1e-5)
        must_link_contig_cov = must_link_contig_cov.apply(lambda x: x + 1e-5)
        contig_cov = contig_cov / 100
        must_link_contig_cov = must_link_contig_cov / 100
        with atomic_write(os.path.join(out, '{}_data_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            contig_cov.to_csv(ofile)

        with atomic_write(os.path.join(out, '{}_data_split_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            must_link_contig_cov.to_csv(ofile)
    else:
        contig_cov = calculate_coverage(
            bam_depth,
            threshold,
            is_combined=is_combined,
            contig_threshold=contig_threshold)
        contig_cov = contig_cov.apply(lambda x: x + 1e-5)
        with atomic_write(os.path.join(out, '{}_data_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            contig_cov.to_csv(ofile)
    return (bam_file, logger)


def generate_cov_multiple(bam_file, bam_index, out, threshold,
                          is_combined, sep, binned_threshold_dict, logger):
    logger.info('Processing {}'.format(bam_file))
    bam_name = os.path.split(bam_file)[-1] + '_{}'.format(bam_index)
    bam_depth = os.path.join(out, '{}_depth.txt'.format(bam_name))
    os.system(
        'bedtools genomecov -bga -ibam {0} > {1}'.format(bam_file, bam_depth))
    if is_combined:
        contig_cov, must_link_contig_cov = calculate_coverage(bam_depth, threshold, is_combined=is_combined,
                                                              sep=sep, binned_thre_dict=binned_threshold_dict)
        contig_cov = contig_cov.apply(lambda x: x + 1e-5)
        must_link_contig_cov = must_link_contig_cov.apply(lambda x: x + 1e-5)
        contig_cov = contig_cov / 100
        must_link_contig_cov = must_link_contig_cov / 100
        with atomic_write(os.path.join(out, '{}_data_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            contig_cov.to_csv(ofile)

        with atomic_write(os.path.join(out, '{}_data_split_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            must_link_contig_cov.to_csv(ofile)
    else:
        contig_cov = calculate_coverage(bam_depth, threshold, is_combined=is_combined, sep=sep,
                                        binned_thre_dict=binned_threshold_dict)
        contig_cov = contig_cov.apply(lambda x: x + 1e-5)
        with atomic_write(os.path.join(out, '{}_data_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            contig_cov.to_csv(ofile)
    return (bam_file, logger)


def _checkback(msg):
    msg[1].info('Processed:{}'.format(msg[0]))


def predict_taxonomy(contig_fasta, GTDB_reference,
                     cannot_name, logger,
                     output, handle,
                     binned_short, must_link_threshold):
    """
    Predict taxonomy using mmseqs and generate cannot-link file
    :param handle: handle to read fasta file
    :param binned_short: binning contigs > 1000bp or 2500 bp
    :param must_link_threshold: threshold for generating must-link pair
    """
    GTDB_default = os.path.join(
        os.environ['HOME'],
        '.cache',
        'S3N2Bin',
        'mmseqs2-GTDB',
        'GTDB')
    GTDB_path = GTDB_reference
    if GTDB_reference is None:
        if not os.path.exists(GTDB_default):
            logger.info('Downloading GTDB.')
            GTDB_dir = os.path.split(GTDB_default)[0]
            os.makedirs(GTDB_dir, exist_ok=True)
            subprocess.check_call(
                ['mmseqs',
                 'databases',
                 'GTDB',
                 GTDB_default,
                 '{}/tmp'.format(GTDB_dir),
                 ],
                stdout=None,
                stderr=subprocess.DEVNULL,
            )
        GTDB_path = GTDB_default
    subprocess.check_call(
        ['mmseqs',
         'createdb',
         contig_fasta,
         '{}/contig_DB'.format(output)],
        stdout=None,
        stderr=subprocess.DEVNULL,
    )
    os.makedirs(os.path.join(output, 'mmseqs_annotation'), exist_ok=True)
    subprocess.run(
        ['mmseqs',
         'taxonomy',
         '{}/contig_DB'.format(output),
         GTDB_path,
         os.path.join(output, 'mmseqs_annotation/mmseqs_annotation'),
         os.path.join(output, 'mmseqs_tmp'),
         '--tax-lineage',
         str(1),
         ],
        check=True,
        stdout=None,
        stderr=subprocess.DEVNULL,
    )
    subprocess.check_call(
        ['mmseqs',
         'createtsv',
         '{}/contig_DB'.format(output),
         os.path.join(output, 'mmseqs_annotation/mmseqs_annotation'),
         os.path.join(output, 'mmseqs_annotation/taxonomyResult.tsv')
         ],
        stdout=None,
        stderr=subprocess.DEVNULL,
    )

    namelist = []
    num_must_link = 0
    binned_threshold = 1000 if binned_short else 2500
    for seq_record in SeqIO.parse(handle, "fasta"):
        if len(seq_record) > binned_threshold:
            namelist.append(seq_record.id)
        if len(seq_record) >= must_link_threshold:
            num_must_link += 1
    os.makedirs(os.path.join(output, 'cannot'), exist_ok=True)
    generate_cannot_link(
        os.path.join(output, 'mmseqs_annotation/taxonomyResult.tsv'),
        namelist, num_must_link,
        os.path.join(output, 'cannot'), cannot_name)


def generate_data_single(bams, num_process, logger,
                         output, handle, binned_short,
                         must_link_threshold):
    """
    Generate data.csv and data_split.csv for training and clustering.
    data.csv has the features(kmer and abundance) for original contigs.
    data_split.csv has the features(kmer and abundace) for contigs that are breaked up as must-link pair.
    """
    n_sample = len(bams)
    is_combined = n_sample >= 5
    bam_list = bams
    if num_process != 0:
        pool = multiprocessing.Pool(num_process)
    else:
        pool = multiprocessing.Pool()

    logger.info('Calculating coverage for every sample.')

    for bam_index in range(n_sample):
        pool.apply_async(
            generate_cov,
            args=(
                bam_list[bam_index],
                bam_index,
                output,
                must_link_threshold,
                is_combined,
                1000 if binned_short else 2500,
                logger,
            ),
            callback=_checkback)
    pool.close()
    pool.join()

    logger.info('Start generating kmer features from fasta file.')
    kmer_whole = generate_kmer_features_from_fasta(
        handle, 1000 if binned_short else 2500, 4)
    kmer_split = generate_kmer_features_from_fasta(
        handle, 1000, 4, split=True, threshold=must_link_threshold)

    data = kmer_whole
    data_split = kmer_split
    data.index = data.index.astype(str)

    for bam_index, bam_file in enumerate(bam_list):
        cov = pd.read_csv(os.path.join(output, '{}_data_cov.csv'.format(
            os.path.split(bam_file)[-1] + '_{}'.format(bam_index))), index_col=0)
        cov.index = cov.index.astype(str)
        data = pd.merge(data, cov, how='inner', on=None,
                        left_index=True, right_index=True, sort=False, copy=True)
        if is_combined:
            cov_split = pd.read_csv(os.path.join(output, '{}_data_split_cov.csv'.format(
                os.path.split(bam_file)[-1] + '_{}'.format(bam_index))), index_col=0)

            data_split = pd.merge(data_split, cov_split, how='inner', on=None,
                                  left_index=True, right_index=True, sort=False, copy=True)
    if not is_combined:
        data_split = kmer_split

    with atomic_write(os.path.join(output, 'data.csv'), overwrite=True) as ofile:
        data.to_csv(ofile)

    with atomic_write(os.path.join(output, 'data_split.csv'), overwrite=True) as ofile:
        data_split.to_csv(ofile)


def generate_data_multi(bams, num_process,separator,
                        logger, output, handle):
    n_sample = len(bams)
    is_combined = n_sample >= 5
    bam_list = bams
    if num_process != 0:
        pool = multiprocessing.Pool(num_process)
    else:
        pool = multiprocessing.Pool()

    # Gererate contig file for every sample
    from collections import defaultdict
    sample_list = list()
    contig_sample_list = []
    contig_length_list = []
    flag_name = None

    os.makedirs(os.path.join(output, 'samples'), exist_ok=True)

    for seq_record in SeqIO.parse(handle, "fasta"):
        sample_name, contig_name = seq_record.id.split(separator)
        if flag_name is None:
            flag_name = sample_name
        if sample_name == flag_name:
            rec = SeqRecord(seq_record.seq, id=contig_name, description='')
            contig_sample_list.append(rec)
        if sample_name != flag_name:
            SeqIO.write(contig_sample_list,
                        os.path.join(
                            output, 'samples', '{}.fasta'.format(flag_name)), 'fasta')
            flag_name = sample_name
            contig_sample_list = []
            rec = SeqRecord(seq_record.seq, id=contig_name, description='')
            contig_sample_list.append(rec)
        if sample_name not in sample_list:
            sample_list.append(sample_name)
        contig_length_list.append(len(seq_record))

    if contig_sample_list != []:
        SeqIO.write(contig_sample_list,
                    os.path.join(
                        output, 'samples', '{}.fasta'.format(flag_name)), 'fasta')

    must_link_threshold = get_threshold(contig_length_list)
    logger.info('Calculating coverage for every sample.')

    binning_threshold = {1000: [], 2500: []}
    for sample in sample_list:
        whole_contig_bp = 0
        contig_bp_2500 = 0
        for seq_record in SeqIO.parse(os.path.join(
                output, 'samples/{}.fasta'.format(sample)), "fasta"):
            if len(seq_record) >= 1000 and len(seq_record) <= 2500:
                contig_bp_2500 += len(seq_record)
            whole_contig_bp += len(seq_record)
        binned_short = contig_bp_2500 / whole_contig_bp < 0.05
        binning_threshold[1000].append(sample) if binned_short \
            else binning_threshold[2500].append(sample)

    for bam_index in range(n_sample):
        pool.apply_async(generate_cov_multiple,
                         args=(
                             bam_list[bam_index],
                             bam_index,
                             os.path.join(output, 'samples'),
                             must_link_threshold,
                             is_combined,
                             separator,
                             binning_threshold,
                             logger
                         ),
                         callback=_checkback)
    pool.close()
    pool.join()
    # Generate cov features for every sample
    data_cov = pd.read_csv(os.path.join(output, 'samples', '{}_data_cov.csv'.format(
        os.path.split(bam_list[0])[-1] + '_{}'.format(0))), index_col=0)
    if is_combined:
        data_split_cov = pd.read_csv(os.path.join(
            output, 'samples', '{}_data_split_cov.csv'.format(
                os.path.split(bam_list[0])[-1] + '_{}'.format(0))), index_col=0)
    data_cov.index = data_cov.index.astype(str)
    for bam_index, bam_file in enumerate(bam_list):
        if bam_index == 0:
            continue
        cov = pd.read_csv(
            os.path.join(output, 'samples', '{}_data_cov.csv'.format(
                os.path.split(bam_file)[-1] + '_{}'.format(bam_index))),
            index_col=0)
        cov.index = cov.index.astype(str)
        data_cov = pd.merge(data_cov, cov, how='inner', on=None,
                            left_index=True, right_index=True, sort=False, copy=True)

        if is_combined:
            cov_split = pd.read_csv(os.path.join(output, 'samples', '{}_data_split_cov.csv'.format(
                os.path.split(bam_file)[-1] + '_{}'.format(bam_index))), index_col=0)

            data_split_cov = pd.merge(data_split_cov, cov_split, how='inner', on=None,
                                      left_index=True, right_index=True, sort=False, copy=True)

    data_cov = data_cov.reset_index()
    columns_list = list(data_cov.columns)
    columns_list[0] = 'contig_name'
    data_cov.columns = columns_list

    if is_combined:
        data_split_cov = data_split_cov.reset_index()
        columns_list = list(data_split_cov.columns)
        columns_list[0] = 'contig_name'
        data_split_cov.columns = columns_list

    for sample in sample_list:
        output_path = os.path.join(output, 'samples', sample)
        os.makedirs(output_path, exist_ok=True)
        part_data = data_cov[data_cov['contig_name'].str.contains(
            '{}'.format(sample + separator))]
        part_data = part_data.set_index('contig_name')
        part_data.index.name = None
        index_list = part_data.index.tolist()
        index_list = [temp.split(separator)[1] for temp in index_list]
        part_data.index = index_list
        part_data.to_csv(os.path.join(output_path, 'data_cov.csv'))
        if is_combined:
            part_data = data_split_cov[data_split_cov['contig_name'].str.contains(
                '{}'.format(sample + separator))]
            part_data = part_data.set_index('contig_name')
            part_data.index.name = None
            index_list = part_data.index.tolist()
            index_list = [temp.split(separator)[1]
                          for temp in index_list]
            part_data.index = index_list
            part_data.to_csv(os.path.join(
                output_path, 'data_split_cov.csv'))

        sample_contig_fasta = os.path.join(
            output, 'samples/{}.fasta'.format(sample))
        binned_short = True if sample in binning_threshold[1000] else False
        kmer_whole = generate_kmer_features_from_fasta(
            sample_contig_fasta, 1000 if binned_short else 2500, 4)
        kmer_split = generate_kmer_features_from_fasta(
            sample_contig_fasta, 1000, 4, split=True, threshold=must_link_threshold)


        sample_cov = pd.read_csv(os.path.join(
            output_path, 'data_cov.csv'),
            index_col=0)
        kmer_whole.index = kmer_whole.index.astype(str)
        sample_cov.index = sample_cov.index.astype(str)
        data = pd.merge(kmer_whole, sample_cov, how='inner', on=None,
                        left_index=True, right_index=True, sort=False, copy=True)
        if is_combined:
            sample_cov_split = pd.read_csv(os.path.join(
                output_path, 'data_split_cov.csv'), index_col=0)
            data_split = pd.merge(kmer_split, sample_cov_split, how='inner', on=None,
                                  left_index=True, right_index=True, sort=False, copy=True)
        else:
            data_split = kmer_split

        with atomic_write(os.path.join(output_path, 'data.csv'), overwrite=True) as ofile:
            data.to_csv(ofile)

        with atomic_write(os.path.join(output_path, 'data_split.csv'), overwrite=True) as ofile:
            data_split.to_csv(ofile)

    return sample_list


def binning(contig_fasta, bams, num_process, data,
            data_split, cannot_link, batchsize, epoches,
            max_edges, max_node, minfasta, logger, output,
            binned_short, device, contig_length_dict, contig_dict):
    """
    Training and clustering the contigs to get the final bins.
    """
    logger.info('Start binning.')
    n_sample = len(bams)
    is_combined = n_sample >= 5
    num_cpu = multiprocessing.cpu_count() if num_process == 0 else num_process
    data = pd.read_csv(data, index_col=0)
    data.index = data.index.astype(str)
    data_split = pd.read_csv(data_split, index_col=0)

    model = train(
        output,
        contig_fasta,
        binned_short,
        logger,
        data,
        data_split,
        cannot_link,
        is_combined,
        batchsize,
        epoches,
        device,
        num_cpu)
    cluster(
        model,
        data,
        device,
        max_edges,
        max_node,
        is_combined,
        logger,
        n_sample,
        contig_length_dict,
        output,
        contig_dict,
        binned_short,
        num_cpu,
        minfasta)


def single_easy_binning(args, logger, output, handle, binned_short,
                        must_link_threshold, device, contig_length_dict, contig_dict):
    logger.info('Running mmseqs and generate cannot-link file.')
    predict_taxonomy(
        args.contig_fasta,
        args.GTDB_reference,
        args.cannot_name,
        logger,
        output, handle, binned_short, must_link_threshold)
    logger.info('Generate training data.')
    generate_data_single(
        args.bams,
        args.num_process,
        logger,
        output, handle, binned_short, must_link_threshold)
    logger.info('Training model and clustering.')
    data_path = os.path.join(output,'data.csv')
    data_split_path = os.path.join(output,'data_split.csv')
    binning(args.contig_fasta, args.bams, args.num_process, data_path,
            data_split_path, os.path.join(
                output, 'cannot', 'cannot.txt'), args.batchsize, args.epoches,
            args.max_edges, args.max_node, args.minfasta_kb * 1000, logger, output, binned_short, device, contig_length_dict, contig_dict)


def multi_easy_binning(args, logger, output, handle, device):
    logger.info('Multi-samples binning.')
    logger.info('Generate training data.')
    sample_list = generate_data_multi(
        args.bams,
        args.num_process,
        args.separator,
        logger,
        output, handle)

    for sample in sample_list:
        logger.info(
            'Running mmseqs and generate cannot-link file of {}.'.format(sample))
        sample_fasta = os.path.join(
            output, 'samples', '{}.fasta'.format(sample))
        sample_data = os.path.join(output, 'samples', sample, 'data.csv')
        sample_data_split = os.path.join(
            output, 'samples', sample, 'data_split.csv')

        whole_contig_bp = 0
        contig_bp_2500 = 0
        contig_length_list = []
        contig_length_dict = {}
        contig_dict = {}
        handle = sample_fasta
        for seq_record in SeqIO.parse(handle, "fasta"):
            if len(seq_record) >= 1000 and len(seq_record) <= 2500:
                contig_bp_2500 += len(seq_record)
            contig_length_list.append(len(seq_record))
            whole_contig_bp += len(seq_record)
            contig_length_dict[str(seq_record.id).strip(
                '')] = len((seq_record.seq))
            contig_dict[str(seq_record.id).strip('')] = str(seq_record.seq)

        binned_short = contig_bp_2500 / whole_contig_bp < 0.05
        must_link_threshold = get_threshold(contig_length_list)

        predict_taxonomy(
            sample_fasta,
            args.GTDB_reference,
            sample,
            logger,
            os.path.join(output, 'samples', sample), handle, binned_short, must_link_threshold,)
        sample_cannot = os.path.join(
            output, 'samples', sample, 'cannot/{}.txt'.format(sample))
        logger.info('Training model and clustering for {}.'.format(sample))
        binning(sample_fasta, args.bams, args.num_process, sample_data,
                sample_data_split, sample_cannot, args.batchsize, args.epoches,
                args.max_edges, args.max_node, args.minfasta_kb * 1000, logger, os.path.join(output, 'samples', sample), binned_short, device, contig_length_dict, contig_dict)

    os.makedirs(os.path.join(output, 'bins'), exist_ok=True)
    for sample in sample_list:
        bin_file = os.listdir(os.path.join(
            output, 'samples', sample, 'output_recluster_bins'))
        for bin in bin_file:
            original_path = os.path.join(
                output, 'samples', sample, 'output_recluster_bins', bin)
            new_file = '{0}_{1}'.format(sample, bin)
            new_path = os.path.join(output, 'bins', new_file)
            shutil.copyfile(original_path, new_path)


def main():
    args = sys.argv[1:]
    args = parse_args(args)

    logger = logging.getLogger('S3N2Bin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    validate_args(args)

    out = args.output
    os.makedirs(out, exist_ok=True)

    device = torch.device(
        "cuda" if torch.cuda.is_available() else "cpu")

    if os.path.splitext(args.contig_fasta)[1] == '.gz':
        handle = gzip.open(args.contig_fasta, "rt")
    elif os.path.splitext(args.contig_fasta)[1] == '.bz2':
        handle = bz2.open(args.contig_fasta, "rt")
    else:
        handle = args.contig_fasta

    if args.cmd in ['predict_taxonomy', 'generate_data_single', 'bin','single_easy_bin']:
        whole_contig_bp = 0
        contig_bp_2500 = 0
        contig_length_list = []
        contig_length_dict = {}
        contig_dict = {}

        for seq_record in SeqIO.parse(handle, "fasta"):
            if len(seq_record) >= 1000 and len(seq_record) <= 2500:
                contig_bp_2500 += len(seq_record)
            contig_length_list.append(len(seq_record))
            whole_contig_bp += len(seq_record)
            contig_length_dict[str(seq_record.id).strip(
                '')] = len((seq_record.seq))
            contig_dict[str(seq_record.id).strip('')] = str(seq_record.seq)

        binned_short = contig_bp_2500 / whole_contig_bp < 0.05
        must_link_threshold = get_threshold(contig_length_list)

    if args.cmd == 'predict_taxonomy':
        predict_taxonomy(
            args.contig_fasta,
            args.GTDB_reference,
            args.cannot_name,
            logger,
            out,
            handle, binned_short, must_link_threshold)

    if args.cmd == 'generate_data_single':
        generate_data_single(
            args.bams,
            args.num_process,
            logger,
            out,
            handle, binned_short, must_link_threshold)

    if args.cmd == 'generate_data_multi':
        generate_data_multi(
            args.bams,
            args.num_process,
            args.separator,
            logger,
            out,
            handle)

    if args.cmd == 'bin':
        binning(args.contig_fasta, args.bams,
                args.num_process, args.data,
                args.data_split, args.cannot_link[0],
                args.batchsize, args.epoches,
                args.max_edges, args.max_node,
                args.minfasta_kb * 1000, logger,
                out, binned_short, device,
                contig_length_dict, contig_dict)

    if args.cmd == 'single_easy_bin':
        single_easy_binning(
            args,
            logger,
            out,
            handle,
            binned_short,
            must_link_threshold,
            device,
            contig_length_dict,
            contig_dict)

    if args.cmd == 'multi_easy_bin':
        multi_easy_binning(
            args,
            logger,
            out,
            handle,
            device)


if __name__ == '__main__':
    main()



import numpy as np
import os
import subprocess
from atomicwrites import atomic_write
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import sys
import pandas as pd
import random

def validate_args(args):

    def expect_file(f):
        if f is not None:
            if not os.path.exists(f):
                sys.stderr.write(
                    f"Error: Expected file '{f}' does not exist\n")
                sys.exit(1)

    def expect_file_list(fs):
        if fs is not None:
            for f in fs:
                expect_file(f)

    if args.cmd == 'predict_taxonomy':
        if args.GTDB_reference is not None:
            expect_file(args.GTDB_reference)
        expect_file(args.contig_fasta)

    if args.cmd == 'generate_data_single':
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

    if args.cmd == 'generate_data_multi':
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

    if args.cmd == 'bin':
        expect_file(args.contig_fasta)
        expect_file(args.data)
        expect_file(args.data_split)

    if args.cmd == 'single_easy_bin':
        if args.GTDB_reference is not None:
            expect_file(args.GTDB_reference)
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

    if args.cmd == 'multi_easy_bin':
        if args.GTDB_reference is not None:
            expect_file(args.GTDB_reference)
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

def get_threshold(contig_len):
    """
    calculate the threshold length for must link breaking up
    """
    basepair_sum = 0
    threshold = 0
    whole_len = np.sum(contig_len)
    contig_len.sort(reverse=True)
    index = 0
    while(basepair_sum / whole_len < 0.98):
        basepair_sum += contig_len[index]
        threshold = contig_len[index]
        index += 1
    return np.clip(threshold, 4000, None)


def parse_mmseqs(mmseqs_result):
    species_result = mmseqs_result.query('rank_name == "species" and score > 0.95').values
    genus_result = mmseqs_result.query('rank_name == "genus" and score > 0.80').values

    cannot_link_species = []
    for i in range(len(species_result)):
        for j in range(i + 1, len(species_result)):
            if species_result[i][2] != species_result[j][2]:
                cannot_link_species.append(
                    (species_result[i][0], species_result[j][0]))

    cannot_link_genus = []
    for i in range(len(genus_result)):
        for j in range(i + 1, len(genus_result)):
            if genus_result[i][2] != genus_result[j][2]:
                cannot_link_genus.append(
                    (genus_result[i][0], genus_result[j][0]))

    cannot_link_mix = []
    for i in range(len(species_result)):
        genus_name = species_result[i][4].split(';')[-2]
        for j in range(len(genus_result)):
            if genus_name != genus_result[j][2]:
                cannot_link_mix.append(
                    (species_result[i][0], genus_result[j][0]))

    return cannot_link_species, cannot_link_genus, cannot_link_mix


def generate_cannot_link(mmseqs_path,namelist,num_threshold,output,sample):
    import itertools
    mmseqs_result = pd.read_csv(mmseqs_path, sep='\t', header=None)
    mmseqs_result.columns = ['contig_name', 'taxon_ID', 'rank_name', 'scientific_name', 'temp_1', 'temp_2', 'temp_3',
                             'score', 'lineage']
    mmseqs_result = mmseqs_result[[
        'contig_name', 'rank_name', 'scientific_name', 'score', 'lineage']]
    mmseqs_result = mmseqs_result[mmseqs_result['contig_name'].isin(
        namelist)]
    cannot_link_species, cannot_link_genus, cannot_link_mix = parse_mmseqs(
        mmseqs_result)
    num_whole_data = np.clip(1000 * num_threshold, None, 4_000_000)
    max_num_mix = int(num_whole_data / 8)
    max_num_genus = int(max_num_mix / 2)
    max_num_species = num_whole_data - max_num_mix - max_num_genus
    if len(cannot_link_mix) > max_num_mix:
        cannot_link_mix = random.sample(cannot_link_mix, max_num_mix)

    if len(cannot_link_genus) > max_num_genus:
        cannot_link_genus = random.sample(cannot_link_genus, max_num_genus)

    if len(cannot_link_species) > max_num_species:
        cannot_link_species = random.sample(cannot_link_species, max_num_species)

    with open(os.path.join(output, f'{sample}.txt'), 'w') as out_cannot_link:
        for cannot in itertools.chain(
                cannot_link_species,
                cannot_link_genus,
                cannot_link_mix):
            out_cannot_link.write(f'{cannot[0]},{cannot[1]}\n')

def cal_num_bins(fasta_path, contig_output, hmm_output,
                 seed_output, binned_short,num_process):
    if not os.path.exists(contig_output + '.faa'):
        frag_out_log = open(contig_output + '.out', 'w')
        subprocess.check_call(
            ['run_FragGeneScan.pl',
             '-genome={}'.format(fasta_path),
             '-out={}'.format(contig_output),
             '-complete=0',
             '-train=complete',
             '-thread={}'.format(num_process),
             ],
            stdout=frag_out_log,
            stderr=subprocess.DEVNULL,
        )

    if not os.path.exists(hmm_output):
        hmm_out_log = open(hmm_output + '.out', 'w')
        subprocess.check_call(
            ['hmmsearch',
             '--domtblout',
             hmm_output,
             '--cut_tc',
             '--cpu', str(num_process),
             os.path.split(__file__)[0] + '/marker.hmm',
             contig_output + '.faa',
             ],
            stdout=hmm_out_log,
            stderr=subprocess.DEVNULL,
        )

    if not os.path.exists(seed_output):
        getmarker = os.path.split(__file__)[0] + '/test_getmarker.pl'
        subprocess.check_call(
            ['perl', getmarker,
             hmm_output,
             fasta_path,
             ('1001' if binned_short else '2501'), # threshold
             seed_output,
             ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )


def write_bins(namelist, contig_labels, output, contig_dict,
               recluster=False, origin_label=0,minfasta = 200000):
    from collections import defaultdict
    res = defaultdict(list)
    for label, name in zip(contig_labels, namelist):
        if label != -1:
            res[label].append(name)

    os.makedirs(output, exist_ok=True)

    for label in res:
        bin = []
        whole_bin_bp = 0
        for contig in res[label]:
            rec = SeqRecord(
                Seq(str(contig_dict[contig])), id=contig, description='')
            bin.append(rec)
            whole_bin_bp += len(str(contig_dict[contig]))

        ofname = f'bin.{label}.fa' if not recluster \
                    else f'recluster_{origin_label}.bin.{label}.fa'
        if whole_bin_bp >= minfasta:
            with atomic_write(os.path.join(output, ofname), overwrite=True) as ofile:
                SeqIO.write(bin, ofile, 'fasta')


def cal_kl(m1, m2, v1, v2):
    m1 = np.clip(m1, 1e-6, None)
    m2 = np.clip(m2, 1e-6, None)
    v1 = np.clip(v1, 1.0, None)
    v2 = np.clip(v2, 1.0, None)
    value = np.log(np.sqrt(v2 / v1)) + \
        np.divide(np.add(v1, np.square(m1 - m2)), 2 * v2) - 0.5
    return np.clip(value, 1e-6, 1 - 1e-6)


import torch
from sklearn.neighbors import kneighbors_graph
from igraph import Graph
import numpy as np
from .utils import cal_kl, write_bins, cal_num_bins
import os
import math
from Bio import SeqIO
from sklearn.cluster import KMeans
import shutil


def cluster(model, data, device, max_edges, max_node, is_combined,
            logger, n_sample, contig_length_dict, out, contig_dict, binned_short,num_process,minfasta):
    train_data = data.values
    if not is_combined:
        train_data_input = train_data[:, 0:136]
    else:
        train_data_input = train_data

    depth = data.values[:, 136:len(data.values[0])]
    namelist = data.index.tolist()
    row_index = data._stat_axis.values.tolist()
    mapObj = {n:i for i, n in enumerate(namelist)}
    with torch.no_grad():
        model.eval()
        x = torch.from_numpy(train_data_input).to(device)
        embedding = model.embedding(x.float()).detach().cpu().numpy()
        embedding_matrix = kneighbors_graph(
            embedding,
            n_neighbors=max_edges,
            mode='distance',
            p=2,
            n_jobs=-1).toarray()
        kmer_matrix = kneighbors_graph(
            train_data_input,
            n_neighbors=max_edges,
            mode='distance',
            p=2,
            n_jobs=-1).toarray()
        embedding_matrix[kmer_matrix == 0] = 0

    embedding_matrix[embedding_matrix >= 1] = 1
    embedding_matrix[embedding_matrix == 0] = 1
    embedding_matrix = 1 - embedding_matrix

    threshold = 0.95

    while (threshold >= 0):
        temp_threshold = embedding_matrix.copy()
        temp_threshold[temp_threshold <= threshold] = 0
        num = len(list(set(np.where(temp_threshold > 0)[0])))
        if round(num / len(embedding_matrix), 2) >= max_node:
            break
        else:
            threshold -= 0.05

    embedding_matrix[embedding_matrix <= threshold] = 0
    if not is_combined:
        logger.info('Calculating depth matrix.')
        depth_matrix = np.zeros(shape=embedding_matrix.shape)
        for i in range(len(embedding_matrix)):
            for j in range(i + 1, len(embedding_matrix)):
                if embedding_matrix[i][j] > 0:
                    temp_depth = 0
                    for k in range(n_sample):
                        temp_depth += 1 - \
                            cal_kl(depth[i][2 * k], depth[j][2 * k],
                                   depth[i][2 * k + 1], depth[j][2 * k + 1])
                    depth_matrix[i][j] = temp_depth / n_sample

        matrix = embedding_matrix * depth_matrix
    else:
        matrix = embedding_matrix

    edges = []
    edges_weight = []

    for i in range(len(matrix)):
        for j in range(i + 1, len(matrix)):
            if matrix[i][j] > 1e-6:
                edges.append((i, j))
                edges_weight.append(matrix[i][j])

    logger.info('Edges:{}'.format(len(edges)))

    g = Graph()
    g.add_vertices(list(range(len(matrix))))
    g.add_edges(edges)
    length_weight = np.array([contig_length_dict[name] for name in namelist])
    result = g.community_infomap(
        edge_weights=edges_weight,
        vertex_weights=length_weight)
    contig_labels = np.zeros(shape=(len(matrix)), dtype=np.int)

    for i, r in enumerate(result):
        for infomap_index in r:
            contig_labels[infomap_index] = i

    output_bin_path = os.path.join(out, 'output_bins')
    os.makedirs(output_bin_path, exist_ok=True)

    write_bins(namelist, contig_labels, output_bin_path, contig_dict,minfasta=minfasta)
    if not is_combined:
        mean_index = [2 * temp for temp in range(n_sample)]
        depth_mean = depth[:, mean_index] / 100
        scaling = np.mean(np.abs(embedding)) / np.mean(depth_mean)
        base = 10
        weight = 2 * base * math.ceil(scaling / base)
        embedding_new = np.concatenate(
            (embedding, depth_mean * weight), axis=1)
    else:
        embedding_new = embedding

    bin_files = os.listdir(output_bin_path)
    logger.info('Reclustering.')

    for bin in bin_files:
        if os.path.exists(os.path.join(output_bin_path, bin)):
            contig_list = []
            for seq_record in SeqIO.parse(
                    os.path.join(output_bin_path, bin), "fasta"):
                contig_list.append(seq_record.id)
            contig_output = os.path.join(output_bin_path, bin) + '.frag'
            hmm_output = os.path.join(output_bin_path, bin) + '.hmmout'
            seed_output = os.path.join(output_bin_path, bin) + '.seed'
            try:
                cal_num_bins(
                    os.path.join(output_bin_path,bin),
                    contig_output,
                    hmm_output,
                    seed_output,
                    binned_short,num_process)
            except BaseException:
                pass
            contig_index = [mapObj[temp] for temp in contig_list]
            re_bin_features = embedding_new[contig_index]
            if not os.path.exists(os.path.join(out, 'output_recluster_bins')):
                os.mkdir(os.path.join(out, 'output_recluster_bins'))

            if os.path.exists(seed_output):
                seed = open(seed_output).read().split('\n')
                seed = [contig for contig in seed if contig != '']
                init_seed = seed
                num_bin = len(seed)
                seed_index = []
                for temp in init_seed:
                    seed_index.append(row_index.index(temp))
                length_weight = np.array(
                    [contig_length_dict[name] for name in contig_list])
                seeds_embedding = embedding_new[seed_index]
                kmeans = KMeans(
                    n_clusters=num_bin,
                    init=seeds_embedding,
                    n_init=1)
                kmeans.fit(re_bin_features, sample_weight=length_weight)
                labels = kmeans.labels_
                write_bins(contig_list, labels, os.path.join(out, 'output_recluster_bins'), contig_dict,
                           recluster=True, origin_label=int(bin.split('.')[-2]),minfasta = minfasta)
            else:
                shutil.copy(os.path.join(
                    output_bin_path, bin), os.path.join(out, 'output_recluster_bins'))




from s3n2bin.main import generate_data_multi
import os
import pytest
import logging
import pandas as pd

def test_generate_data_multi():
    logger = logging.getLogger('S3N2Bin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    os.makedirs('output_multi',exist_ok=True)
    generate_data_multi(bams=['multi_samples_data/input_multi_sorted1.bam',
                              'multi_samples_data/input_multi_sorted2.bam',
                              'multi_samples_data/input_multi_sorted3.bam',
                              'multi_samples_data/input_multi_sorted4.bam',
                              'multi_samples_data/input_multi_sorted5.bam',
                              'multi_samples_data/input_multi_sorted6.bam',
                              'multi_samples_data/input_multi_sorted7.bam',
                              'multi_samples_data/input_multi_sorted8.bam',
                              'multi_samples_data/input_multi_sorted9.bam',
                              'multi_samples_data/input_multi_sorted10.bam'],
                         num_process=1,
                         separator=':',
                         logger=logger,
                         output='output_multi',
                         handle='multi_samples_data/input_multi.fasta',
                         )

    for i in range(10):
        data = pd.read_csv('output_multi/samples/S{}/data.csv'.format(i+1),index_col=0)
        data_split = pd.read_csv('output_multi/samples/S{}/data_split.csv'.format(i+1),index_col=0)

        assert data.shape == (20,146)
        assert data_split.shape == (40,146)



from s3n2bin.main import generate_data_single
import os
import pytest
import logging
import pandas as pd

def test_generate_data_coassembly():
    logger = logging.getLogger('S3N2Bin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    os.makedirs('output_coassembly',exist_ok=True)
    generate_data_single(bams=['coassembly_sample_data/input.sorted1.bam','coassembly_sample_data/input.sorted2.bam','coassembly_sample_data/input.sorted3.bam','coassembly_sample_data/input.sorted4.bam','coassembly_sample_data/input.sorted5.bam'],
                         num_process=1,
                         logger=logger,
                         output='output_coassembly',
                         handle='coassembly_sample_data/input.fasta',
                         binned_short=False,
                         must_link_threshold=4000
                         )

    data = pd.read_csv('output_coassembly/data.csv',index_col=0)
    data_split = pd.read_csv('output_coassembly/data_split.csv',index_col=0)

    assert data.shape == (40,141)
    assert data_split.shape == (80,141)


"""
https://github.com/BinPro/CONCOCT/blob/develop/scripts/fasta_to_features.py
"""

import numpy as np
from itertools import product
from Bio import SeqIO
from itertools import tee
from collections import Counter, OrderedDict
import pandas as pd
from Bio.SeqRecord import SeqRecord


def window(seq, n):
    els = tee(seq, n)
    for i, el in enumerate(els):
        for _ in range(i):
            next(el, None)
    return zip(*els)


def generate_feature_mapping(kmer_len):
    BASE_COMPLEMENT = {"A": "T", "T": "A", "G": "C", "C": "G"}
    kmer_hash = {}
    counter = 0
    for kmer in product("ATGC", repeat=kmer_len):
        if kmer not in kmer_hash:
            kmer_hash[kmer] = counter
            rev_compl = tuple([BASE_COMPLEMENT[x] for x in reversed(kmer)])
            kmer_hash[rev_compl] = counter
            counter += 1
    return kmer_hash, counter


def generate_kmer_features_from_fasta(
        fasta_file, length_threshold, kmer_len, split=False, threshold=0):
    kmer_dict, nr_features = generate_feature_mapping(kmer_len)
    composition_d = OrderedDict()
    contig_lengths = OrderedDict()
    if not split:
        seq_list = list(SeqIO.parse(fasta_file, "fasta"))
    else:
        seq_list = []
        for seq_record in SeqIO.parse(fasta_file, "fasta"):
            if len(seq_record) >= threshold:
                half = int(len(seq_record.seq) / 2)
                rec1 = SeqRecord(
                    seq_record.seq[0:half], id=seq_record.id + '_1', description='')
                seq_list.append(rec1)
                rec2 = SeqRecord(seq_record.seq[half:len(
                    seq_record.seq)], id=seq_record.id + '_2', description='')
                seq_list.append(rec2)
    for seq in seq_list:
        seq_len = len(seq)
        if seq_len <= length_threshold:
            continue
        contig_lengths[seq.id] = seq_len
        kmers = [
            kmer_dict[kmer_tuple]
            for kmer_tuple
            in window(str(seq.seq).upper(), kmer_len)
            if kmer_tuple in kmer_dict
        ]
        kmers.append(nr_features - 1)
        composition_v = np.bincount(np.array(kmers, dtype=np.int64))
        composition_v[-1] -= 1
        composition_d[seq.id] = composition_v
    df = pd.DataFrame.from_dict(composition_d, orient='index', dtype=float)

    df = df.apply(lambda x: x + 1e-5)
    df = df.div(df.sum(axis=1), axis=0)
    return df



import torch
from torch.nn import Linear, ReLU, LeakyReLU
from torch import nn
from torch.utils.data import Dataset, DataLoader
import os
from .utils import cal_num_bins
import pandas as pd
from torch.optim import lr_scheduler
from tqdm import tqdm


class Semi_encoding_multiple(torch.nn.Module):
    def __init__(self, num):
        super(Semi_encoding_multiple, self).__init__()
        self.encoder1 = torch.nn.Sequential(
            Linear(num, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 100),
        )

        self.decoder1 = torch.nn.Sequential(
            Linear(100, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, num),
            nn.Sigmoid(),
        )

    def forward(self, input1, input2):
        return self.encoder1(input1), self.encoder1(input2)

    def decoder(self, input1, input2):
        return self.decoder1(input1), self.decoder1(input2)

    def embedding(self, input):
        return self.encoder1(input)


class Semi_encoding_single(torch.nn.Module):
    def __init__(self, num):
        super(Semi_encoding_single, self).__init__()
        self.encoder1 = torch.nn.Sequential(
            Linear(num, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 100),
        )

        self.decoder1 = torch.nn.Sequential(
            Linear(100, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, num),
            nn.Softmax(dim=1),
        )

    def forward(self, input1, input2):
        return self.encoder1(input1), self.encoder1(input2)

    def decoder(self, input1, input2):
        return self.decoder1(input1), self.decoder1(input2)

    def embedding(self, input):
        return self.encoder1(input)


def loss_function(embedding1, embedding2, label, raw_x_1,
                  raw_x_2, decoder_x_1, decoder_x_2, is_label=True):
    relu = torch.nn.ReLU()
    mse_loss = torch.nn.MSELoss()
    d = torch.norm(embedding1 - embedding2, p=2, dim=1)
    square_pred = torch.square(d)
    margin_square = torch.square(relu(1 - d))

    if is_label:
        supervised_loss = torch.mean(
            label * square_pred + (1 - label) * margin_square)
        unsupervised_loss = 0.5 * \
            mse_loss(decoder_x_1, raw_x_1) + 0.5 * \
            mse_loss(decoder_x_2, raw_x_2)
        loss = supervised_loss + unsupervised_loss
        return loss, supervised_loss, unsupervised_loss

    else:
        unsupervised_loss = 0.5 * \
            mse_loss(decoder_x_1, raw_x_1) + 0.5 * \
            mse_loss(decoder_x_2, raw_x_2)
        return unsupervised_loss


class feature_Dataset(Dataset):

    def __init__(self, embedding1_lists, embedding2_lists, label_lists):
        self.embedding1_lists = embedding1_lists
        self.embedding2_lists = embedding2_lists
        self.label_lists = label_lists

    def __getitem__(self, item):
        return self.embedding1_lists[item], self.embedding2_lists[item], self.label_lists[item]

    def __len__(self):
        return len(self.embedding1_lists)


class unsupervised_feature_Dataset(Dataset):
    def __init__(self, embedding1_lists, embedding2_lists):
        self.embedding1_lists = embedding1_lists
        self.embedding2_lists = embedding2_lists

    def __getitem__(self, item):
        return self.embedding1_lists[item], self.embedding2_lists[item]

    def __len__(self):
        return len(self.embedding1_lists)


def train(out, contig_fasta, binned_short, logger, data, data_split, cannot_link, is_combined=True,
          batchsize=2048, epoches=20, device=None,num_process = 8):
    contig_output = os.path.join(out, os.path.split(contig_fasta)[1] + '.frag')
    hmm_output = os.path.join(out, os.path.split(contig_fasta)[1] + '.hmmout')
    seed_output = os.path.join(out, os.path.split(contig_fasta)[1] + '.seed')


    cal_num_bins(
        contig_fasta,
        contig_output,
        hmm_output,
        seed_output,
        binned_short,
        num_process)



    logger.info('Generate training data:')

    # #generate two inputs
    train_input_1 = []
    train_input_2 = []
    train_labels = []

    # cannot link
    cannot_link = pd.read_csv(cannot_link, sep=',', header=None).values

    namelist = data.index.tolist()
    mapObj = dict(zip(namelist, range(len(namelist))))
    train_data = data.values
    train_data_must_link = data_split.values

    if not is_combined:
        train_data_input = train_data[:, 0:136]
        train_data_split_input = train_data_must_link
    else:
        train_data_input = train_data
        train_data_split_input = train_data_must_link

    # can not link from contig annotation
    for link in cannot_link:
        train_input_1.append(train_data_input[mapObj[str(link[0])]])
        train_input_2.append(train_data_input[mapObj[str(link[1])]])
        train_labels.append(0)

    # cannot link from bin seed
    if os.path.exists(seed_output):
        seed = open(seed_output).read().split('\n')
        seed = [contig for contig in seed if contig != '']
        for i in range(len(seed)):
            for j in range(i + 1, len(seed)):
                train_input_1.append(train_data_input[mapObj[str(seed[i])]])
                train_input_2.append(train_data_input[mapObj[str(seed[j])]])
                train_labels.append(0)

    # must link from breaking up
    for i in range(0, len(train_data_must_link), 2):
        train_input_1.append(train_data_split_input[i])
        train_input_2.append(train_data_split_input[i + 1])
        train_labels.append(1)

    logger.info('Number of must link pair:{}'.format(train_labels.count(1)))
    logger.info('Number of can not link pair:{}'.format(train_labels.count(0)))

    logger.info('Training model...')

    dataset = feature_Dataset(train_input_1, train_input_2, train_labels)
    train_loader = DataLoader(
        dataset=dataset,
        batch_size=batchsize,
        shuffle=True,
        num_workers=0)

    unlabeled_x = train_data_input
    unlabeled_train_input1 = []
    unlabeled_train_input2 = []
    for i in range(len(unlabeled_x)):
        unlabeled_train_input1.append(unlabeled_x[i])
        unlabeled_train_input2.append(unlabeled_x[i])

    dataset_unlabeled = unsupervised_feature_Dataset(
        unlabeled_train_input1, unlabeled_train_input2)
    train_loader_unlabeled = DataLoader(
        dataset=dataset_unlabeled,
        batch_size=batchsize,
        shuffle=True,
        num_workers=16)

    if not is_combined:
        model = Semi_encoding_single(train_data_input.shape[1]).to(device)
    else:
        model = Semi_encoding_multiple(train_data_input.shape[1]).to(device)
    optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
    scheduler = lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.9)

    for epoch in tqdm(range(epoches)):
        for train_input1, train_input2, train_label in train_loader:
            model.train()
            train_input1 = train_input1.to(device)
            train_input2 = train_input2.to(device)
            train_label = train_label.to(device)
            embedding1, embedding2 = model.forward(
                train_input1.float(), train_input2.float())
            decoder1, decoder2 = model.decoder(embedding1, embedding2)
            optimizer.zero_grad()
            loss, supervised_loss, unsupervised_loss = loss_function(embedding1.double(), embedding2.double(),
                                                                     train_label.double(), train_input1.double(),
                                                                     train_input2.double(), decoder1.double(),
                                                                     decoder2.double())
            loss = loss.to(device)
            loss.backward()
            optimizer.step()

        for unlabeled_train_input1, unlabeled_train_input2 in train_loader_unlabeled:
            model.train()
            unlabeled_train_input1 = unlabeled_train_input1.to(device)
            unlabeled_train_input2 = unlabeled_train_input2.to(device)
            embedding1, embedding2 = model.forward(
                unlabeled_train_input1.float(), unlabeled_train_input1.float())
            decoder1, decoder2 = model.decoder(embedding1, embedding2)
            optimizer.zero_grad()
            loss = loss_function(embedding1.double(), embedding2.double(),
                                 None, unlabeled_train_input1.double(),
                                 unlabeled_train_input2.double(), decoder1.double(),
                                 decoder2.double(), is_label=False).to(device)
            loss.backward()
            optimizer.step()
        scheduler.step()

    logger.info('Training finished.')
    torch.save(model, os.path.join(out, 'model.h5'))

    return model


from s3n2bin.main import binning
import os
import pytest
import logging
import pandas as pd
from Bio import SeqIO

def test_bin():
    logger = logging.getLogger('S3N2Bin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    contig_length_dict = {}
    contig_dict = {}
    handle = 'bin_data/input.fasta'
    for seq_record in SeqIO.parse(handle, "fasta"):
        contig_length_dict[str(seq_record.id).strip(
            '')] = len((seq_record.seq))
        contig_dict[str(seq_record.id).strip('')] = str(seq_record.seq)

    os.makedirs('output_bin',exist_ok=True)
    binning(contig_fasta='bin_data/input.fasta',
            bams=['bin_data/input.sorted.bam'],
            num_process=1,
            data='bin_data/data.csv',
            data_split='bin_data/data_split.csv',
            cannot_link='bin_data/cannot.txt',
            batchsize=2048,
            epoches=1,
            max_edges=20,
            max_node=1,
            minfasta=0,
            logger=logger,
            output='output_bin',
            binned_short=True,
            device='cpu',
            contig_length_dict=contig_length_dict,
            contig_dict=contig_dict
            )

    assert len(os.listdir('output_bin/output_bins')) > 0
    assert len(os.listdir('output_bin/output_recluster_bins')) > 0


from s3n2bin.generate_coverage import calculate_coverage
from os import path
import pytest
import pandas as pd
from pandas.testing import assert_frame_equal


def test_cov(tmpdir):
    depth_file = path.join(tmpdir, 'depth.txt')

    with open(depth_file, 'wt') as out_depth:
        out_depth.write('k141_63080\t0\t6\t1\n')
        out_depth.write('k141_63080\t6\t30\t2\n')
        out_depth.write('k141_63080\t30\t138\t3\n')
        out_depth.write('k141_63080\t138\t144\t2\n')
        out_depth.write('k141_63080\t144\t174\t1\n')
        out_depth.write('k141_63080\t174\t177\t2\n')
        out_depth.write('k141_63080\t177\t257\t3\n')
        out_depth.write('k141_63080\t257\t304\t2\n')
        out_depth.write('k141_63080\t304\t315\t1\n')
        out_depth.write('k141_0\t0\t45\t1\n')
        out_depth.write('k141_0\t45\t88\t2\n')
        out_depth.write('k141_0\t88\t138\t3\n')
        out_depth.write('k141_0\t138\t159\t2\n')
        out_depth.write('k141_0\t159\t183\t3\n')
        out_depth.write('k141_0\t183\t225\t2\n')
        out_depth.write('k141_0\t225\t227\t1\n')
        out_depth.write('k141_0\t227\t297\t3\n')
        out_depth.write('k141_0\t297\t338\t2\n')
        out_depth.write('k141_0\t338\t365\t1\n')

    cov, cov_split = calculate_coverage(
        depth_file,
        threshold=0,
        is_combined=True,
        contig_threshold=0)
    assert_frame_equal(cov, pd.DataFrame([2.581818, 2.627907], index=[
                       'k141_63080', 'k141_0'], columns=['cov']))
    assert_frame_equal(cov_split, pd.DataFrame([2.609756, 2.554217, 2.682243, 2.574074], index=[
                       'k141_63080_1', 'k141_63080_2', 'k141_0_1', 'k141_0_2'], columns=['cov']))

    cov = calculate_coverage(
        depth_file,
        threshold=0,
        is_combined=False,
        contig_threshold=0)
    assert_frame_equal(cov, pd.DataFrame([[2.581818, 0.606942], [
                       2.627907, 0.252244]], index=['k141_63080', 'k141_0'], columns=['mean', 'var']))
